/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

The algorithms found in this code are based on publicly released source from
Scott Scriven.

******************************************************************************

###CLASS###
Name:      Reflection
Version:   1.0
Status:    Beta
Date:      July 2003
ID:        ID_REFLECTION
Category:  Effect
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2003.  All rights reserved.
Short:     Creates water reflections in real time.
Keywords:  water, reflection, mirror, wave

###DESCRIPTION###
<p>The Reflection class draws an animated effect that emulates the appearance of
reflected water.  Although it may appear complex, the reflection algorithm uses
a simple line copying technique, at very little cost to the CPU.  This makes it an
excellent effect to use when an animation is required that isn't going to
be a burden on the system.</p>

<p>Once initialised, the effect will run automatically until the object is
freed from the system.</p>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/input/keyboard.h>
#include <pandora/input/joyport.h>
#include <pandora/graphics/bitmap.h>
#include <pandora/graphics/render.h>
#include <pandora/system/module.h>
#include <math.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_REFLECTION 1.0

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_REFLECTION, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2003.  All rights reserved.",
   "July 2003",
   "Reflection"
};

static struct KernelBase *KernelBase = 0;
static OBJECTPTR ReflectionClass = 0;

/*****************************************************************************
** Reflection class definition.
*/

struct Reflection {
   OBJECT_HEADER
   LONG  Speed;
   FLOAT TopIntensity;
   FLOAT BottomIntensity;
   FLOAT WaterLevel;

   PRIVATE_FIELDS
   LONG  Opacity;
   OBJECTID DrawableID;    /* Reference to drawable destination */
   LONG  Buttons;           /* Storage for held mouse buttons */
   LONG  Dimensions;
   LONG  DrawableWidth;
   LONG  DrawableHeight;
   FLOAT XCoord;
   FLOAT YCoord;
   FLOAT YOffset;
   FLOAT XOffset;
   FLOAT Width;
   FLOAT Height;
   FLOAT prvSpeedCounter;
};

static ERROR GET_Height(struct Reflection *, struct Variable *);
static ERROR GET_Opacity(struct Reflection *, FLOAT *);
static ERROR GET_Width(struct Reflection *, struct Variable *);
static ERROR GET_X(struct Reflection *, struct Variable *);
static ERROR GET_XOffset(struct Reflection *, struct Variable *);
static ERROR GET_Y(struct Reflection *, struct Variable *);
static ERROR GET_YOffset(struct Reflection *, struct Variable *);

static ERROR SET_Height(struct Reflection *, struct Variable *);
static ERROR SET_Opacity(struct Reflection *, FLOAT *);
static ERROR SET_Speed(struct Reflection *, LONG *);
static ERROR SET_WaterLevel(struct Reflection *, FLOAT *);
static ERROR SET_Width(struct Reflection *, struct Variable *);
static ERROR SET_X(struct Reflection *, struct Variable *);
static ERROR SET_XOffset(struct Reflection *, struct Variable *);
static ERROR SET_Y(struct Reflection *, struct Variable *);
static ERROR SET_YOffset(struct Reflection *, struct Variable *);

static struct FieldArray ReflectionFields[] = {
   { "Speed",           0, FDF_LONG|FDF_RW,      0, NULL, SET_Speed },
   { "TopIntensity",    0, FDF_FLOAT|FDF_RW,     0, NULL, NULL },
   { "BottomIntensity", 0, FDF_FLOAT|FDF_RW,     0, NULL, NULL },
   { "WaterLevel",      0, FDF_FLOAT|FDF_RW,     0, NULL, SET_WaterLevel },
   /*** Variable Fields ***/
   { "Opacity",     0, FDF_FLOAT|FDF_PERCENTAGE|FDF_RW,               0, GET_Opacity, SET_Opacity },
   { "Height",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Width",       0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",           0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XCoord",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XOffset",     0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",           0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YCoord",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YOffset",     0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   END_FIELD
};

static ERROR REFLECTION_ActionNotify(struct Reflection *, struct acActionNotify *);
static ERROR REFLECTION_Free(struct Reflection *, APTR);
static ERROR REFLECTION_Init(struct Reflection *, APTR);
static ERROR REFLECTION_NewObject(struct Reflection *, APTR);
static ERROR REFLECTION_Timer(struct Reflection *, struct acTimer *);

static struct ActionArray ReflectionActions[] = {
   { AC_ActionNotify, REFLECTION_ActionNotify },
   { AC_Free,         REFLECTION_Free },
   { AC_Init,         REFLECTION_Init },
   { AC_NewObject,    REFLECTION_NewObject },
   { AC_Timer,        REFLECTION_Timer },
   { NULL, NULL }
};

/****************************************************************************/

#ifndef PI
#define PI 3.14159265358979323846
#define PI2 (PI * 2)
#endif

#define SPEED           24.0 /* The higher the value, the slower we go */
#define START_INTENSITY 14.0 /* High values produce less intense ripples at the top of the effect */
#define END_INTENSITY   28.0 /* High values produce slower ripples at the bottom of the effect */

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&ReflectionClass, NULL,
      FID_BaseClassID|TLONG, ID_REFLECTION,
      FID_Version|TFLOAT,    VER_REFLECTION,
      FID_Name|TSTRING,      "Reflection",
      FID_Category|TLONG,    CCF_EFFECT,
      FID_Actions|TPTR,      ReflectionActions,
      FID_Fields|TPTR,       ReflectionFields,
      FID_Size|TLONG,        sizeof(struct Reflection),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (ReflectionClass) { Action(AC_Free, ReflectionClass, NULL); ReflectionClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Reflection: ActionNotify
*/

static ERROR REFLECTION_ActionNotify(struct Reflection *Self, struct acActionNotify *NotifyArgs)
{
   if ((NotifyArgs->ActionID IS AC_Draw) AND (NotifyArgs->Error IS ERR_Okay)) {
      struct Bitmap *bitmap;
      struct Variable var;
      struct mtBlitArea blit;
      OBJECTPTR drawable;
      FLOAT d, eheight;
      LONG xcoord, ycoord, width, height, drawableheight, visibleheight;
      LONG j, offset, waterheight;
      WORD opacity;

      if (AccessObject(Self->DrawableID, 0, &drawable) != ERR_Okay) {
         return(ObjectError(Self, ERH_Draw, ERR_ExclusiveDenied));
      }

      if ((GetField(drawable, FID_Bitmap, FT_POINTER, &bitmap) != ERR_Okay) OR (!bitmap)) {
         ReleaseObject(drawable);
         return(ObjectError(Self, ERH_Draw, ERR_GetField));
      }

      GetFields(drawable, FID_Height|TLONG, &drawableheight, TAGEND);

      var.Type = FD_LARGE;
      GET_X(Self, &var);      xcoord = var.Large;
      GET_Y(Self, &var);      ycoord = var.Large;
      GET_Width(Self, &var);  width  = var.Large;
      GET_Height(Self, &var); height = var.Large;

      /*** Perform the algorithm ***/

      d = (PI2 * (DOUBLE)Self->prvSpeedCounter) / Self->Speed;
      opacity = bitmap->Opacity;
      bitmap->Opacity = Self->Opacity;

      blit.DestBitmap = bitmap;
      blit.Flags      = NULL;
      blit.XCoord     = xcoord;
      blit.Width      = width;
      blit.Height     = 1;
      blit.XDest      = xcoord;

      waterheight = (height * Self->WaterLevel) / 100;
      visibleheight = height - waterheight;
      eheight = (FLOAT)height / Self->BottomIntensity;

      for (j=ycoord; j <= ycoord + visibleheight; j++) {
         /* Calculate the magic offset, which determines the
         ** source line that we are copying from.
         */

         offset = (eheight * (j + Self->TopIntensity) * sin((eheight * (height - j)) / (j + 1.0) + d) ) / waterheight;

         blit.YCoord = ycoord + (waterheight - j + offset); /* Source Line, modified by offset */
         if (blit.YCoord < 0) blit.YCoord = -blit.YCoord;
         blit.YDest = ycoord + ((waterheight - j) + (j<<1)); /* Destination Line */
         Action(MT_BlitArea, bitmap, &blit);
      }

      /*** Clean up any left-over lines with a little repeat blitting ***/

      while (++blit.YDest < height) Action(MT_BlitArea, bitmap, &blit);

      bitmap->Opacity = opacity;

      ReleaseObject(drawable);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Reflection: Free
*/

static ERROR REFLECTION_Free(struct Reflection *Self, APTR Void)
{
   UnsubscribeTimer(Self->Head.UniqueID);

   /*** Remove all subscriptions to our Drawable container ***/

   if (Self->DrawableID) {
      OBJECTPTR drawable;
      if (AccessObject(Self->DrawableID, 3000, &drawable) IS ERR_Okay) {
         UnsubscribeAction(drawable, NULL,  GetUniqueID(Self));
         ReleaseObject(drawable);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Reflection: Init
*/

static ERROR REFLECTION_Init(struct Reflection *Self, APTR Void)
{
   struct Variable value;
   OBJECTPTR Drawable;

   /*** Find the Drawable object that we are associated with ***/

   if (!Self->DrawableID) {
      OBJECTID OwnerID;
      OwnerID = GetOwner(Self);
      while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
         OwnerID = GetOwnerID(OwnerID);
      }
      if (!OwnerID) return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
      else Self->DrawableID = OwnerID;
   }

   /* If dimension settings are missing (e.g. if it is impossible to
   ** determine width, height or a coordinate), then we set the missing
   ** fields to maximum possible values.
   */

   if (!(Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X|DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
      value.Type = FD_DOUBLE;
      value.Double = 0;
      SET_X(Self, &value);
   }

   if (!(Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y|DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
      value.Type = FD_DOUBLE;
      value.Double = 0;
      SET_Y(Self, &value);
   }

   if (!(Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH|DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
      value.Type = FD_DOUBLE;
      value.Double = 0;
      SET_XOffset(Self, &value);
   }

   if (!(Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT|DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
      value.Type = FD_DOUBLE;
      value.Double = 0;
      SET_YOffset(Self, &value);
   }

   if (AccessObject(Self->DrawableID, 5000, &Drawable) IS ERR_Okay) {
      /*** Subscribe to our container's Resize action ***/

      SubscribeActionTags(Drawable, Self,
         AC_Draw,
         TAGEND);
      ReleaseObject(Drawable);
   }
   else return(ObjectError(Self, ERH_Init, ERR_AccessObject));

   /*** Initialise the Timer ***/

   return(SubscribeTimer(Self->Head.UniqueID, 60));
}

/*****************************************************************************
** Reflection: NewObject
*/

static ERROR REFLECTION_NewObject(struct Reflection *Self, APTR Void)
{
   Self->Opacity         = 255;
   Self->Speed           = SPEED;           /* Speed at which the water shifts */
   Self->TopIntensity    = START_INTENSITY; /* Use low values to intensify the ripples at the top of the effect */
   Self->BottomIntensity = END_INTENSITY;   /* High values produce less intense ripples at the bottom of the effect */
   Self->WaterLevel      = 50;              /* % level at which the water starts */
   return(ERR_Okay);
}

/*****************************************************************************
** Action: Timer
*/

static ERROR REFLECTION_Timer(struct Reflection *Self, struct acTimer *Args)
{
   struct Variable value;
   struct acDraw draw;

   if (++Self->prvSpeedCounter >= Self->Speed) Self->prvSpeedCounter = 0;

   value.Type = FD_DOUBLE;
   GET_X(Self, &value);      draw.XCoord = value.Double;
   GET_Y(Self, &value);      draw.YCoord = value.Double;
   GET_Width(Self, &value);  draw.Width  = value.Double;
   GET_Height(Self, &value); draw.Height = value.Double;
   DelayMsg(AC_Draw, Self->DrawableID, &draw);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Height
Short:  The height of the ripple graphic is specified here.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>If a ripple graphic is to be limited to a specific drawable area, set this
field to limit the clipping height.  A percentage may be specified if the
FD_PERCENT flag is used when setting the field.</p>

###SEE ALSO###
Field: ReflectionHeight, ReflectionWidth, Width

###END###

*****************************************************************************/

static ERROR GET_Height(struct Reflection *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   FLOAT containerheight, YCoord, YOffset;

   if (!(Value->Type & (FD_DOUBLE|FD_LARGE))) {
      return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   }

   if (Value->Type & FD_PERCENTAGE) {
      if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
         Value->Double = Self->Height;
      }
      else if (Self->Dimensions & DMF_FIXED_HEIGHT) {
         if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
            GetField(drawable, FID_Height, FT_FLOAT, &containerheight);
            Value->Double = Self->Height / containerheight * 100;
            ReleaseObject(drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
      }
      else {
         if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
                  (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
            if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
               GetField(drawable, FID_Height, FD_FLOAT, &containerheight);
               if (Self->Dimensions & DMF_FIXED_Y) YCoord = Self->YCoord;
               else YCoord = Self->YCoord * 100 / containerheight;
               if (Self->Dimensions & DMF_FIXED_YOFFSET) YOffset = Self->YOffset;
               else YOffset = Self->YOffset * 100 / containerheight;
               Value->Double = (containerheight - YOffset - YCoord) / containerheight * 100;
               ReleaseObject(drawable);
            }
            else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
         }
         else return(ERR_FieldNotSet);
      }
   }
   else if (Self->Dimensions & DMF_FIXED_HEIGHT) {
      Value->Double = Self->Height;
   }
   else if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
      if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
         GetField(drawable, FID_Height, FT_FLOAT, &containerheight);
         Value->Double = Self->Height * containerheight / 100;
         ReleaseObject(drawable);
      }
      else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
   }
   else {
      if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
            GetField(drawable, FID_Height, FD_FLOAT, &containerheight);
            GetFields(Self, FID_YCoord|TFLOAT,  &YCoord,
                            FID_YOffset|TFLOAT, &YOffset,
                            TAGEND);
            Value->Double = containerheight - YOffset - YCoord;
            ReleaseObject(drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
      }
      else return(ERR_FieldNotSet);
   }

   Value->Large = Value->Double;
   return(ERR_Okay);
}

static ERROR SET_Height(struct Reflection *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Reflection] A height setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_HEIGHT) | DMF_FIXED_HEIGHT;
      Self->Height     = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Reflection] A height setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_HEIGHT) | DMF_RELATIVE_HEIGHT;
      Self->Height     = Value->Double;
   }
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   Opacity
Short:  Determines the level of translucency applied to the reflection.
Type:   FLOAT/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>This field determines the translucency level of the reflected water.  The
default setting is 100%, which means that the reflection will be solid.  Any
other value that you set here will alter the impact of the reflection over
the destination drawable.</p>

<p>Please note that the use of translucency will always have an impact on the
time it normally takes to draw the reflection.</p>

###END###

****************************************************************************/

static ERROR GET_Opacity(struct Reflection *Self, FLOAT *Value)
{
   *Value = Self->Opacity * 100 / 255;
   return(ERR_Okay);
}

static ERROR SET_Opacity(struct Reflection *Self, FLOAT *Value)
{
   if (*Value < 0) Self->Opacity = 0;
   else if (*Value > 100) Self->Opacity = 255;
   else Self->Opacity = *Value * 255 / 100;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Speed
Short:  The speed at which the reflection vertically shifts can be altered in this field.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>To later the speed at which the reflection moves, you can write values
to this field.  Lower values will increase the speed and higher values will
lower it.  The recommended speed setting is 24.</p>

###END###

*****************************************************************************/

static ERROR SET_Speed(struct Reflection *Self, LONG *Value)
{
   if (*Value < 2) Self->Speed = 2;
   else if (*Value > 5000) Self->Speed = 5000;
   else Self->Speed = *Value;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   WaterLevel
Short:  Determines the level at which the water reflection starts.
Type:   FLOAT/PERCENTAGE
Status: Read/Set

###DESCRIPTION###
<p>The level at which the reflected water starts can be altered by writing to
this field.  The default, minimum value is 50%.  Increasing the value will
lower the water level (so a setting of 75% will fill 25% of the graphics space
with water).  The maximum value is 95%.</p>

###END###

*****************************************************************************/

static ERROR SET_WaterLevel(struct Reflection *Self, FLOAT *Value)
{
   if (*Value < 50) Self->WaterLevel = 50;
   else if (*Value > 95) Self->WaterLevel = 95;
   else Self->WaterLevel = *Value;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Width
Short:  The width of the ripple graphic is specified here.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>If a ripple graphic is to be limited to a specific drawable area, set this
field to limit the clipping width.  A percentage may be specified if the
FD_PERCENT flag is used when setting the field.</p>

###SEE ALSO###
Field: Height

###END###

*****************************************************************************/

static ERROR GET_Width(struct Reflection *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   FLOAT ContainerWidth, XCoord, XOffset;

   if (!(Value->Type & (FD_DOUBLE|FD_LARGE))) {
      return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   }

   if (Value->Type & FD_PERCENTAGE) {
      if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
         Value->Double = Self->Width;
      }
      else if (Self->Dimensions & DMF_FIXED_WIDTH) {
         if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
            GetField(drawable, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->Width / ContainerWidth * 100;
            ReleaseObject(drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
      }
      else {
         if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
                  (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
            if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
               GetField(drawable, FID_Width, FD_FLOAT, &ContainerWidth);
               if (Self->Dimensions & DMF_FIXED_X) XCoord = Self->XCoord;
               else XCoord = Self->XCoord * 100 / ContainerWidth;
               if (Self->Dimensions & DMF_FIXED_XOFFSET) XOffset = Self->XOffset;
               else XOffset = Self->XOffset * 100 / ContainerWidth;
               Value->Double = (ContainerWidth - XOffset - XCoord) * 100 / ContainerWidth;
               ReleaseObject(drawable);
            }
            else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
         }
         else return(ERR_FieldNotSet);
      }
   }
   else if (Self->Dimensions & DMF_FIXED_WIDTH) {
      Value->Double = Self->Width;
   }
   else if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
      if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
         GetField(drawable, FID_Width, FT_FLOAT, &ContainerWidth);
         Value->Double = Self->Width * ContainerWidth / 100;
         ReleaseObject(drawable);
      }
      else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
   }
   else {
      if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
            GetField(drawable, FID_Width, FT_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_X) XCoord = Self->XCoord;
            else XCoord = ContainerWidth * Self->XCoord / 100;
            if (Self->Dimensions & DMF_FIXED_XOFFSET) Value->Double = ContainerWidth - XCoord - Self->XOffset;
            else Value->Double = ContainerWidth - XCoord - (ContainerWidth * Self->XOffset / 100);
            ReleaseObject(drawable);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
      }
      else return(ERR_Failed);
   }

   Value->Large = Value->Double;
   return(ERR_Okay);
}

static ERROR SET_Width(struct Reflection *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Reflection] A width setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_WIDTH) | DMF_FIXED_WIDTH;
      Self->Width      = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Reflection] A width setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_WIDTH) | DMF_RELATIVE_WIDTH;
      Self->Width      = Value->Double;
   }
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  Defines the horizontal position of the ripple graphic.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The horizontal position of an image can be set to an absolute or relative
coordinate by writing a value to the XCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Fields: Align, XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct Reflection *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerwidth, width, value;

   if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_X) value = Self->XCoord;
      else if (Self->Dimensions & DMF_RELATIVE_X) {
         value = Self->XCoord * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_WIDTH) width = Self->Width;
         else width = containerwidth * Self->Width / 100;
         if (Self->Dimensions & DMF_FIXED_XOFFSET) value = containerwidth - width - Self->XOffset;
         else value = containerwidth - width - (containerwidth * Self->XOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_X(struct Reflection *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->XCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->XCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_X) | DMF_RELATIVE_X;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_X) | DMF_FIXED_X;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  Defines the horizontal offset of the ripple graphic.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The XOffset has a dual purpose depending on whether or not it is set in
conjunction with an X coordinate or a Width based field.</p>

<p>If set in conjunction with an X coordinate then the image will be drawn
from that X coordinate up to the width of the container, minus the value
given in the XOffset.  This means that the width of the image is
dynamically calculated in relation to the width of the container.</p>

<p>If the XOffset field is set in conjunction with a fixed or relative width
then the image will be drawn at an X coordinate calculated from the formula
"XCoord = ContainerWidth - ReflectionWidth - XOffset".</p>

###SEE ALSO###
Fields: Align, XCoord, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct Reflection *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerwidth, width, value;

   if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_XOFFSET) value = Self->XOffset;
      else if (Self->Dimensions & DMF_RELATIVE_XOFFSET) {
         value = Self->XOffset * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) {
         if (Self->Dimensions & DMF_FIXED_WIDTH) width = Self->Width;
         else width = containerwidth * Self->Width / 100;
         if (Self->Dimensions & DMF_FIXED_X) value = containerwidth - (Self->XCoord + width);
         else value = containerwidth - ((Self->XCoord * containerwidth / 100) + width);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_XOffset(struct Reflection *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->XOffset = Value->Double;
   else if (Value->Type & FD_LARGE) Self->XOffset = Value->Large;
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_XOFFSET) | DMF_RELATIVE_XOFFSET;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_XOFFSET) | DMF_FIXED_XOFFSET;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  Defines the vertical position of the ripple graphic.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The vertical position of an image can be set to an absolute or relative
coordinate by writing a value to the YCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Fields: Align, XCoord, YOffset

###END###

*****************************************************************************/

static ERROR GET_Y(struct Reflection *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value, containerheight, height;

   if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_Y) value = Self->YCoord;
      else if (Self->Dimensions & DMF_RELATIVE_Y) {
         value = Self->YCoord * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_HEIGHT) height = Self->Height;
         else height = containerheight * Self->Height / 100;
         if (Self->Dimensions & DMF_FIXED_YOFFSET) value = containerheight - height - Self->YOffset;
         else value = containerheight - height - (containerheight * Self->YOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Y(struct Reflection *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->YCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->YCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_Y) | DMF_RELATIVE_Y;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_Y) | DMF_FIXED_Y;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  Defines the vertical offset of the ripple graphic.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The YOffset has a dual purpose depending on whether or not it is set in
conjunction with a Y coordinate or a Height based field.</p>

<p>If set in conjunction with a Y coordinate then the image will be drawn
from that Y coordinate up to the height of the container, minus the value
given in the YOffset.  This means that the height of the ripple is
dynamically calculated in relation to the height of the container.</p>

<p>If the YOffset field is set in conjunction with a fixed or relative height
then the image will be drawn at a Y coordinate calculated from the formula
"YCoord = ContainerHeight - Height - YOffset".</p>

###SEE ALSO###
Fields: Align, XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct Reflection *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value, containerheight, height;

   value = 0;
   if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_YOFFSET) value = Self->YOffset;
      else if (Self->Dimensions & DMF_RELATIVE_YOFFSET) {
         value = Self->YOffset * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if (Self->Dimensions & DMF_FIXED_HEIGHT) height = Self->Height;
         else height = containerheight * Self->Height / 100;
         if (Self->Dimensions & DMF_FIXED_Y) value = containerheight - (Self->YCoord + height);
         else value = containerheight - ((Self->YCoord * containerheight / 100) + height);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetDimension, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_YOffset(struct Reflection *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->YOffset= Value->Double;
   else if (Value->Type & FD_LARGE) Self->YOffset= Value->Large;
   else return(ObjectError(Self, ERH_SetDimension, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_YOFFSET) | DMF_RELATIVE_YOFFSET;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_YOFFSET) | DMF_FIXED_YOFFSET;
   return(ERR_Okay);
}
