/*****************************************************************************

This source code and its accompanying files are in the public domain and
therefore may be distributed without restriction.  The source is based in
part on libpng, authored by Glenn Randers-Pehrson, Andreas Eric Dilger and
Guy Eric Schalnat, and zlib, authored by Jean-loup Gailly and Mark Adler.

*****************************************************************************/

#define PNG_INTERNAL
#include "lib/png.h"

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/files/assign.h>
#include <pandora/files/file.h>
#include <pandora/modules/filesystem.h>

#define VER_PNGMod 1.0

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_PNGMod, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2001-2003.  Freely distributable.",
   "April 2003",
   "PNG"
};

static struct KernelBase *KernelBase = 0;
static OBJECTPTR PNGClass = 0;
static OBJECTPTR modFileSystem = 0;
static struct FileSystemBase *FileSystemBase = 0;
static LONG glError = FALSE;

static ERROR PNG_Activate(struct Picture *, APTR);
static ERROR PNG_Init(struct Picture *, APTR);
static ERROR PNG_Query(struct Picture *, APTR);
static ERROR PNG_SaveImage(struct Picture *, struct acSaveImage *);
static ERROR PNG_SaveToObject(struct Picture *, struct acSaveToObject *);

static void read_row_callback(png_structp, png_uint_32, int);
static void write_row_callback(png_structp, png_uint_32, int);

static struct ActionArray PNGActions[] = {
   { AC_Activate,     PNG_Activate },
   { AC_Init,         PNG_Init },
   { AC_Query,        PNG_Query },
   { AC_SaveImage,    PNG_SaveImage },
   { AC_SaveToObject, PNG_SaveToObject },
   { NULL, NULL }
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   /*** File System module ***/

   if (CreateObject(ID_MODULE, NULL, &modFileSystem, NULL, FID_Name|TSTRING, "filesystem", TAGEND) IS ERR_Okay) {
      if (GetField(modFileSystem, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_NewObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&PNGClass, NULL,
      FID_BaseClassID|TLONG,       ID_PICTURE,
      FID_SubClassID|TLONG,        ID_PNG,
      FID_Name|TSTRING,            "PNG",
      FID_Category|TLONG,          CCF_GRAPHICS,
      FID_FileExtension|TSTRING,   "png",
      FID_FileDescription|TSTRING, "PNG Picture",
      FID_Actions|TPTR,            PNGActions,
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (modFileSystem) { acFree(modFileSystem); modFileSystem = NULL; }
   if (PNGClass)      { acFree(PNGClass); PNGClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** PNG: Activate
*/

static ERROR PNG_Activate(struct Picture *Self, APTR Void)
{
   struct Bitmap *Bitmap;
   struct RGB rgb;
   ERROR error;
   UBYTE *row;
   STRING location;
   LONG i, x, y;
   png_bytep row_pointers;
   png_structp read_ptr;
   png_infop info_ptr;
   png_infop end_info;
   png_uint_32 width, height;
   int bit_depth, color_type;
   png_color_16p background;

   DPrintF("~Activate()","[PNG:%d]", Self->Head.UniqueID);

   /*** Return if the picture object has already been activated ***/

   if (Self->Bitmap->Head.ObjectFlags & NF_INITIALISED) return(ERR_Okay);

   /*** Initialise variables ***/

   Bitmap   = Self->Bitmap;
   read_ptr = NULL;
   info_ptr = NULL;
   end_info = NULL;
   error    = ERR_Failed;
   glError  = FALSE;

   /*** Open the data file ***/

   if (!Self->prvFile) {
      if (GetFields(Self, FID_Location|TSTRING, &location, TAGEND) != ERR_Okay) return(ObjectError(Self, ERH_Activate, ERR_GetField));

      if (CreateObject(ID_FILE, NULL, &Self->prvFile, NULL,
            FID_Location|TSTRING, location,
            FID_Flags|TLONG,      FL_READ|FL_APPROXIMATE,
            TAGEND) != ERR_Okay) goto exit;
   }

   ActionTags(AC_Seek, Self->prvFile, 0.0, SEEK_START);

   /*** Allocate PNG structures ***/

   if (!(read_ptr = png_create_read_struct(0, NULL, NULL, NULL))) goto exit;
   if (!(info_ptr = png_create_info_struct(read_ptr))) goto exit;
   if (!(end_info = png_create_info_struct(read_ptr))) goto exit;

   /*** Setup the PNG file ***/

   read_ptr->io_ptr = Self->prvFile;
   read_ptr->read_data_fn = png_read_data;
   read_ptr->output_flush_fn = NULL;

   png_set_read_status_fn(read_ptr, read_row_callback); if (glError) goto exit;
   png_set_invert_alpha(read_ptr); if (glError) goto exit;
   png_read_info(read_ptr, info_ptr); if (glError) goto exit;
   png_get_IHDR(read_ptr, info_ptr, &width, &height, &bit_depth, &color_type, NULL, NULL, NULL); if (glError) goto exit;

   if (!Bitmap->Width)  Bitmap->Width  = width;
   if (!Bitmap->Height) Bitmap->Height = height;
   if (!Bitmap->Type)   Bitmap->Type   = BMP_CHUNKY;

   if (!Self->DisplayWidth)  Self->DisplayWidth  = width;
   if (!Self->DisplayHeight) Self->DisplayHeight = height;

   /*** If the image contains a palette, load the palette into our Bitmap ***/

   if (info_ptr->valid & PNG_INFO_PLTE) {
      for (i=0; (i < info_ptr->num_palette) AND (i < 256); i++) {
         Bitmap->Palette->Col[i].Red   = info_ptr->palette[i].red;
         Bitmap->Palette->Col[i].Green = info_ptr->palette[i].green;
         Bitmap->Palette->Col[i].Blue  = info_ptr->palette[i].blue;
         Bitmap->Palette->Col[i].Alpha = 0;
      }
   }

   /* If the picture supports an alpha channel, initialise
   ** an alpha based Mask object for the Picture.
   */

   if (color_type & PNG_COLOR_MASK_ALPHA) {
      if (Self->Flags & PCF_32BITALPHA) {
         /* Upgrade the image to 32-bit and store the alpha channel
         ** in the alpha byte of the pixel data.
         */
         Bitmap->BitsPerPixel  = 32;
         Bitmap->BytesPerPixel = 4;
         Bitmap->Flags |= BMF_ALPHA;
      }
      else {
         if (NewObject(ID_BITMAP, Self->Head.ObjectFlags|NF_CHILD, &Self->Mask, &Self->prvMaskID) IS ERR_Okay) {
            Self->Mask->Width  = Self->Bitmap->Width;
            Self->Mask->Height = Self->Bitmap->Height;
            Self->Mask->AmtColours = 256;
            Self->Mask->Flags      |= BMF_MASK;
            if (Action(AC_Init, Self->Mask, NULL) IS ERR_Okay) {
               Self->Flags |= PCF_MASK|PCF_ALPHA;
            }
            else goto exit;
         }
      }
   }

   /* If a background colour has been specified for the image (instead
   ** of an alpha channel), read it and create the mask based on the
   ** data that we have read.
   */

   if (info_ptr->valid & PNG_INFO_tRNS) {
      /* The first colour index in the list is taken as the
      ** background, any others are ignored
      */

      if ((info_ptr->color_type IS PNG_COLOR_TYPE_PALETTE) OR
          (info_ptr->color_type IS PNG_COLOR_TYPE_GRAY) OR
          (info_ptr->color_type IS PNG_COLOR_TYPE_GRAY_ALPHA)) {
         Bitmap->ColourIndex = info_ptr->trans[0];
         rgb.Red   = Bitmap->Palette->Col[Bitmap->ColourIndex].Red;
         rgb.Green = Bitmap->Palette->Col[Bitmap->ColourIndex].Green;
         rgb.Blue  = Bitmap->Palette->Col[Bitmap->ColourIndex].Blue;
         rgb.Alpha = 0;
         SetField(Bitmap, FID_ColourRGB, FT_POINTER, &rgb);
         DPrintF("Activate:","[PNG] Transparent Colour: %d,%d,%d [Palette Index %d]", rgb.Red, rgb.Green, rgb.Blue, info_ptr->trans[0]);
      }
      else {
         rgb.Red   = info_ptr->trans_values.red;
         rgb.Green = info_ptr->trans_values.green;
         rgb.Blue  = info_ptr->trans_values.blue;
         rgb.Alpha = 0;
         SetField(Bitmap, FID_ColourRGB, FT_POINTER, &rgb);
         DPrintF("Activate:","[PNG] Transparent Colour: %d,%d,%d", rgb.Red, rgb.Green, rgb.Blue);
      }

      Self->Flags |= PCF_BACKGROUND;
   }

   if (info_ptr->valid & PNG_INFO_bKGD) {
      background = &(info_ptr->background);
      if (color_type IS PNG_COLOR_TYPE_PALETTE) {
         Bitmap->ColourIndex = background->index;
         rgb.Red   = Bitmap->Palette->Col[Bitmap->ColourIndex].Red;
         rgb.Green = Bitmap->Palette->Col[Bitmap->ColourIndex].Green;
         rgb.Blue  = Bitmap->Palette->Col[Bitmap->ColourIndex].Blue;
         rgb.Alpha = 0;
         SetField(Bitmap, FID_ColourRGB, FT_POINTER, &rgb);
         DPrintF("Activate:","[PNG] Background Colour: %d,%d,%d [Palette Index %d]", rgb.Red, rgb.Green, rgb.Blue, background->index);
      }
      else if ((color_type IS PNG_COLOR_TYPE_GRAY) OR (color_type IS PNG_COLOR_TYPE_GRAY_ALPHA)) {
         rgb.Red   = background->gray;
         rgb.Green = background->gray;
         rgb.Blue  = background->gray;
         rgb.Alpha = 0;
         SetField(Bitmap, FID_ColourRGB, FT_POINTER, &rgb);
         DPrintF("Activate:","[PNG] Background Colour: %d,%d,%d", rgb.Red, rgb.Green, rgb.Blue);
      }
      else {
         rgb.Red   = background->red;
         rgb.Green = background->green;
         rgb.Blue  = background->blue;
         rgb.Alpha = 0;
         SetField(Bitmap, FID_ColourRGB, FT_POINTER, &rgb);
         DPrintF("Activate:","[PNG] Background Colour: %d,%d,%d", rgb.Red, rgb.Green, rgb.Blue);
      }

      Self->Flags |= PCF_BACKGROUND;
   }

   /*** Set the bits per pixel value ***/

   if (!Bitmap->BitsPerPixel) {
      if ((color_type IS PNG_COLOR_TYPE_GRAY) OR (color_type IS PNG_COLOR_TYPE_PALETTE)) {
         Bitmap->BitsPerPixel = 8;
      }
      else Bitmap->BitsPerPixel = 24;
   }

   /*** Read the image data into our Bitmap ***/

   if (color_type & PNG_COLOR_MASK_ALPHA) png_set_expand(read_ptr); /* Alpha channel */
   if (bit_depth IS 16) png_set_strip_16(read_ptr); /* Reduce bit depth to 24bpp if the image is 48bpp */
   if (bit_depth < 8) png_set_packing(read_ptr);

   if ((error = Action(AC_Query, Bitmap, NULL)) IS ERR_Okay) {
      if ((error = Action(AC_Init, Bitmap, NULL)) IS ERR_Okay) {
         if ((error = AllocMemory(png_get_rowbytes(read_ptr, info_ptr), MEM_DATA|MEM_NOCLEAR, &row, NULL)) IS ERR_Okay) {

            row_pointers = row;
            if (color_type IS PNG_COLOR_TYPE_GRAY) {
               for (y=0; y < height; y++) {
                  png_read_row(read_ptr, row_pointers, NULL); if (glError) goto exit;
                  for (x=0; x < width; x++) {
                     rgb.Red   = row[x];
                     rgb.Green = row[x];
                     rgb.Blue  = row[x];
                     rgb.Alpha = 0;
                     Bitmap->DrawUCRPixel(Bitmap, x, y, &rgb);
                  }
               }
            }
            else if (color_type IS PNG_COLOR_TYPE_PALETTE) {
               if (Bitmap->BitsPerPixel IS 8) {
                  for (y=0; y < height; y++) {
                     png_read_row(read_ptr, row_pointers, NULL); if (glError) goto exit;
                     for (x=0; x < width; x++) Bitmap->DrawUCPixel(Bitmap, x, y, row[x]);
                  }
               }
               else {
                  for (y=0; y < height; y++) {
                     png_read_row(read_ptr, row_pointers, NULL); if (glError) goto exit;
                     for (x=0; x < width; x++) {
                        rgb.Red   = Bitmap->Palette->Col[row[x]].Red;
                        rgb.Green = Bitmap->Palette->Col[row[x]].Green;
                        rgb.Blue  = Bitmap->Palette->Col[row[x]].Blue;
                        rgb.Alpha = 0;
                        Bitmap->DrawUCRPixel(Bitmap, x, y, &rgb);
                     }
                  }
               }
            }
            else if (color_type & PNG_COLOR_MASK_ALPHA) {
               /* When decompressing images that support an alpha channel, the fourth
               ** byte of each pixel will contain the alpha data.
               */

               for (y=0; y < height; y++) {
                  png_read_row(read_ptr, row_pointers, NULL); if (glError) goto exit;
                  i = 0;
                  for (x=0; x < width; x++) {
                     rgb.Red   = row[i++];
                     rgb.Green = row[i++];
                     rgb.Blue  = row[i++];
                     rgb.Alpha = row[i++];
                     Bitmap->DrawUCRPixel(Bitmap, x, y, &rgb);

                     /* Set the alpha byte in the alpha mask (nb: refer to
                     ** png_set_invert_alpha() if you want to reverse the alpha bytes)
                     */

                     if (Self->Mask) Self->Mask->Data[(y * Self->Mask->LineWidth) + x] = rgb.Alpha;
                  }
               }
            }
            else for (y=0; y < height; y++) {
               png_read_row(read_ptr, row_pointers, NULL);
               if (glError) goto exit;
               i = 0;
               for (x=0; x < width; x++) {
                  rgb.Red   = row[i++];
                  rgb.Green = row[i++];
                  rgb.Blue  = row[i++];
                  rgb.Alpha = 0;
                  Bitmap->DrawUCRPixel(Bitmap, x, y, &rgb);
               }
            }

            FreeMemory(row);
            png_read_end(read_ptr, end_info);
         }
         else goto exit;
      }
      else goto exit;
   }
   else goto exit;

   if (Self->prvFile) {
      acFree(Self->prvFile);
      Self->prvFile = NULL;
   }

   if (Self->Flags & PCF_CACHE) Action(MT_CacheBitmap, Self, NULL);

   error = ERR_Okay;

exit:
   if (error) ObjectError(Self, ERH_Activate, error);

   png_destroy_read_struct(&read_ptr, &info_ptr, &end_info);
   StepBack();
   return(error);
}

/*****************************************************************************
** PNG: Init
*/

static ERROR PNG_Init(struct Picture *Self, APTR Void)
{
   UBYTE *buffer;
   STRING location;

   GetField(Self, FID_Location, FT_STRING, &location);

   if ((!location) OR (Self->Flags & PCF_NEW)) {
      /* If no location has been specified, assume that the picture is being
      ** created from scratch (e.g. to save an image to disk).  The
      ** programmer is required to specify the dimensions and colours of the
      ** Bitmap so that we can initialise it.
      */

      if (!Self->Bitmap->Width) Self->Bitmap->Width = Self->DisplayWidth;
      if (!Self->Bitmap->Height) Self->Bitmap->Height = Self->DisplayHeight;

      if ((Self->Bitmap->Width) AND (Self->Bitmap->Height)) {
         if (Action(AC_Init, Self->Bitmap, NULL) IS ERR_Okay) {

            if (Self->Flags & PCF_32BITALPHA) Self->Flags &= ~(PCF_ALPHA|PCF_MASK);

            if (Self->Flags & PCF_ALPHA) {
               if (NewObject(ID_BITMAP, Self->Head.ObjectFlags|NF_CHILD, &Self->Mask, &Self->prvMaskID) IS ERR_Okay) {
                  Self->Mask->Width  = Self->Bitmap->Width;
                  Self->Mask->Height = Self->Bitmap->Height;
                  Self->Mask->Flags |= PCF_MASK;
                  Self->Mask->BitsPerPixel = 8;
                  if (acInit(Self->Mask) IS ERR_Okay) Self->Flags |= PCF_MASK;
                  else return(ObjectError(Self, ERH_Init, ERR_Init));
               }
               else return(ObjectError(Self, ERH_Init, ERR_NewObject));
            }
            else if (Self->Flags & PCF_MASK) {
               if (NewObject(ID_BITMAP, Self->Head.ObjectFlags|NF_CHILD, &Self->Mask, &Self->prvMaskID) IS ERR_Okay) {
                  Self->Mask->Width  = Self->Bitmap->Width;
                  Self->Mask->Height = Self->Bitmap->Height;
                  Self->Mask->Flags |= PCF_MASK;
                  Self->Mask->BitsPerPixel = 1;
                  if (acInit(Self->Mask) IS ERR_Okay) Self->Flags |= PCF_MASK;
                  else return(ObjectError(Self, ERH_Init, ERR_Init));
               }
               else return(ObjectError(Self, ERH_Init, ERR_NewObject));
            }

            return(ERR_Okay);
         }
         else return(ObjectError(Self, ERH_Init, ERR_Init));
      }
      else {
         DPrintF("@PNG:","The Bitmap width/height fields have not been set.");
         return(ERR_FieldNotSet);
      }
   }
   else if (GetField(Self, FID_Header, FT_POINTER, &buffer) IS ERR_Okay) {
      if ((buffer[0] IS 0x89) AND (buffer[1] IS 0x50) AND (buffer[2] IS 0x4e) AND (buffer[3] IS 0x47) AND
          (buffer[4] IS 0x0d) AND (buffer[5] IS 0x0a) AND (buffer[6] IS 0x1a) AND (buffer[7] IS 0x0a)) {
         if (!(Self->Flags & PCF_QUERY)) Action(AC_Activate, Self, NULL);
         return(ERR_Okay);
      }
   }

   return(ERR_NoSupport);
}

/*****************************************************************************
** PNG: Query
*/

static ERROR PNG_Query(struct Picture *Self, APTR Void)
{
   struct Bitmap *Bitmap;
   png_structp read_ptr;
   png_infop info_ptr;
   png_infop end_info;
   STRING location;
   png_uint_32 width, height;
   int bit_depth, color_type;
   ERROR error;

   Bitmap   = Self->Bitmap;
   error    = ERR_Failed;
   read_ptr = NULL;
   info_ptr = NULL;
   end_info = NULL;
   glError  = FALSE;

   /*** Open the data file ***/

   if (!Self->prvFile) {
      if (GetFields(Self, FID_Location|TSTRING, &location, TAGEND) != ERR_Okay) return(ObjectError(Self, ERH_Query, ERR_GetField));

      if (CreateObject(ID_FILE, NULL, &Self->prvFile, NULL,
            FID_Location|TSTRING, location,
            FID_Flags|TLONG,      FL_READ|FL_APPROXIMATE,
            TAGEND) != ERR_Okay) goto exit;
   }

   ActionTags(AC_Seek, Self->prvFile, 0.0, SEEK_START);

   /*** Allocate PNG structures ***/

   if (!(read_ptr = png_create_read_struct(0, NULL, NULL, NULL))) goto exit;
   if (!(info_ptr = png_create_info_struct(read_ptr))) goto exit;
   if (!(end_info = png_create_info_struct(read_ptr))) goto exit;

   /*** Read the PNG description ***/

   read_ptr->io_ptr = Self->prvFile;
   read_ptr->read_data_fn = png_read_data;
   read_ptr->output_flush_fn = NULL;

   png_set_read_status_fn(read_ptr, read_row_callback); if (glError) goto exit;
   png_set_invert_alpha(read_ptr); if (glError) goto exit;
   png_read_info(read_ptr, info_ptr); if (glError) goto exit;
   png_get_IHDR(read_ptr, info_ptr, &width, &height, &bit_depth, &color_type, NULL, NULL, NULL); if (glError) goto exit;

   if (!Bitmap->Width)  Bitmap->Width  = width;
   if (!Bitmap->Height) Bitmap->Height = height;
   if (!Bitmap->Type)   Bitmap->Type   = BMP_CHUNKY;

   if (!Self->DisplayWidth)  Self->DisplayWidth  = width;
   if (!Self->DisplayHeight) Self->DisplayHeight = height;
   if (color_type & PNG_COLOR_MASK_ALPHA) Self->Flags |= PCF_ALPHA;

   if (!Bitmap->BitsPerPixel) {
      if ((color_type IS PNG_COLOR_TYPE_GRAY) OR (color_type IS PNG_COLOR_TYPE_PALETTE)) {
         Bitmap->BitsPerPixel = 8;
         Bitmap->BytesPerPixel = 1;
      }
      else {
         Bitmap->BitsPerPixel = 24;
         Bitmap->BytesPerPixel = 3;
      }
   }

   Action(AC_Query, Bitmap, NULL);

   error = ERR_Okay;

exit:
   png_destroy_read_struct(&read_ptr, &info_ptr, &end_info);
   return(error);
}

/*****************************************************************************
** PNG: SaveImage
*/

static ERROR PNG_SaveImage(struct Picture *Self, struct acSaveImage *Args)
{
   struct Bitmap *Bitmap;
   OBJECTPTR file;
   ERROR error;
   STRING location;
   LONG y, i, x;
   png_bytep row_pointers;
   png_structp write_ptr;
   png_infop info_ptr;
   png_color_16 background;

   DPrintF("~SaveImage()","[PNG:%d]", Self->Head.UniqueID);

   /*** Initialise variables ***/

   Bitmap    = Self->Bitmap;
   file      = NULL;
   write_ptr = NULL;
   info_ptr  = NULL;
   error     = ERR_Failed;
   glError   = FALSE;

   /*** Open the data file ***/

   if ((Args) AND (Args->DestinationID)) {
      if (AccessObject(Args->DestinationID, 3000, &file) != ERR_Okay) {
         DPrintF("@SaveImage:","[PNG] Failed to access destination object #%d.", Args->DestinationID);
         goto exit;
      }
   }
   else {
      if (GetFields(Self, FID_Location|TSTRING, &location, TAGEND) != ERR_Okay) {
         ObjectError(Self, ERH_SaveImage, ERR_MissingLocation);
         StepBack();
         return(ERR_GetField);
      }

      if (CreateObject(ID_FILE, NULL, &file, NULL,
            FID_Location|TSTRING, location,
            FID_Flags|TLONG,      FL_NEW|FL_WRITE,
            TAGEND) != ERR_Okay) {
         ObjectError(Self, ERH_SaveImage, ERR_CreateObject);
         StepBack();
         return(ERR_CreateObject);
      }
   }

   /*** Allocate PNG structures ***/

   if (!(write_ptr = png_create_write_struct(0, NULL, NULL, NULL))) {
      DPrintF("@SaveImage:","[PNG] png_create_write_struct() failed.");
      goto exit;
   }

   if (!(info_ptr = png_create_info_struct(write_ptr))) {
      DPrintF("@SaveImage:","[PNG] png_create_info_struct() failed.");
      goto exit;
   }

   /*** Setup the PNG file ***/

   write_ptr->io_ptr = file;
   write_ptr->write_data_fn = png_write_data;
   write_ptr->output_flush_fn = NULL;

   png_set_write_status_fn(write_ptr, write_row_callback);
   if (glError) {
      DPrintF("@SaveImage:","[PNG] png_set_write_status_fn() failed.");
      goto exit;
   }

   png_set_invert_alpha(write_ptr);
   if (glError) {
      DPrintF("@SaveImage:","[PNG] png_set_invert_alpha() failed.");
      goto exit;
   }

   if ((Self->Flags & PCF_32BITALPHA) AND (Self->Bitmap->BitsPerPixel != 32)) {
      DPrintF("@SaveImage:","[PNG] Illegal use of the 32BITALPHA flag in a %d bit bitmap.", Self->Bitmap->BitsPerPixel);
      Self->Flags &= ~PCF_32BITALPHA;
   }

   if ((Self->Flags & (PCF_ALPHA|PCF_MASK)) AND (!Self->Mask)) {
      DPrintF("@SaveImage:","[PNG] Illegal use of the ALPHA/MASK flags without an accompanying mask bitmap.");
      Self->Flags &= ~(PCF_ALPHA|PCF_MASK);
   }

   if (Self->Bitmap->AmtColours > 256) {
      if (Self->Flags & PCF_32BITALPHA) {
         png_set_IHDR(write_ptr, info_ptr, Self->Bitmap->Width, Self->Bitmap->Height, 8, PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
      }
      else if (Self->Flags & PCF_ALPHA) {
         png_set_IHDR(write_ptr, info_ptr, Self->Bitmap->Width, Self->Bitmap->Height, 8, PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
      }
      else png_set_IHDR(write_ptr, info_ptr, Self->Bitmap->Width, Self->Bitmap->Height, 8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
   }
   else {
      png_set_IHDR(write_ptr, info_ptr, Self->Bitmap->Width, Self->Bitmap->Height, 8, PNG_COLOR_TYPE_PALETTE, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
      png_set_PLTE(write_ptr, info_ptr, (png_colorp)Self->Bitmap->Palette->Col, Self->Bitmap->AmtColours);
   }

   /*** On Intel CPU's the pixel format is BGR ***/

   png_set_bgr(write_ptr);

   /*** Set the transparent colour if PCF_BACKGROUND is used ***/

   if (Self->Flags & PCF_BACKGROUND) {
      background.index = Self->Bitmap->ColourIndex;
      background.red   = Self->Bitmap->ColourRGB->Red;
      background.green = Self->Bitmap->ColourRGB->Green;
      background.blue  = Self->Bitmap->ColourRGB->Blue;
      png_set_bKGD(write_ptr, info_ptr, &background);
   }

   /*** Write the header to the PNG file ***/

   png_write_info(write_ptr, info_ptr);
   if (glError) {
      DPrintF("@SaveImage:","[PNG] png_write_info() failed.");
      goto exit;
   }

   /*** Write the image data to the PNG file ***/

   if ((Self->Bitmap->BitsPerPixel IS 8) OR (Self->Bitmap->BitsPerPixel IS 24)) {
      if (Self->Flags & PCF_ALPHA) {
         UBYTE row[Self->Bitmap->Width * 4];
         UBYTE *data, *mask;
         WORD maskx;

         row_pointers = row;
         data = Self->Bitmap->Data;
         mask = Self->Mask->Data;
         for (y=0; y < Self->Bitmap->Height; y++) {
            i = 0;
            maskx = 0;
            for (x=0; x < Self->Bitmap->ByteWidth; x+=3) {
               row[i++] = data[x+0];  /* Blue */
               row[i++] = data[x+1];  /* Green */
               row[i++] = data[x+2];  /* Red */
               row[i++] = mask[maskx++];  /* Alpha */
            }
            png_write_row(write_ptr, row_pointers);
            data += Self->Bitmap->LineWidth;
            mask += Self->Mask->LineWidth;
         }
      }
      else {
         for (y=0; y < Self->Bitmap->Height; y++) {
            row_pointers = Self->Bitmap->Data + (y * Self->Bitmap->LineWidth);
            png_write_row(write_ptr, row_pointers);
         }
      }
   }
   else if (Self->Bitmap->BitsPerPixel IS 32) {
      UBYTE *data, *mask;

      if (Self->Flags & PCF_32BITALPHA) {
         UBYTE row[Self->Bitmap->Width * 4];

         row_pointers = row;
         data = Self->Bitmap->Data;
         for (y=0; y < Self->Bitmap->Height; y++) {
            i = 0;
            for (x=0; x < (Self->Bitmap->Width<<2); x+=4) {
               row[i++] = data[x+0];  /* Blue */
               row[i++] = data[x+1];  /* Green */
               row[i++] = data[x+2];  /* Red */
               row[i++] = data[x+3];  /* Alpha */
            }
            png_write_row(write_ptr, row_pointers);
            data += Self->Bitmap->LineWidth;
         }
      }
      else if (Self->Flags & PCF_ALPHA) {
         UBYTE row[Self->Bitmap->Width * 4];
         WORD maskx;

         row_pointers = row;
         data = Self->Bitmap->Data;
         mask = Self->Mask->Data;
         for (y=0; y < Self->Bitmap->Height; y++) {
            i = 0;
            maskx = 0;
            for (x=0; x < (Self->Bitmap->Width<<2); x+=4) {
               row[i++] = data[x+0];     /* Blue */
               row[i++] = data[x+1];     /* Green */
               row[i++] = data[x+2];     /* Red */
               row[i++] = mask[maskx++]; /* Alpha */
            }
            png_write_row(write_ptr, row_pointers);
            data += Self->Bitmap->LineWidth;
            mask += Self->Mask->LineWidth;
         }
      }
      else {
         UBYTE row[Self->Bitmap->Width * 3];
         row_pointers = row;
         data = Self->Bitmap->Data;
         for (y=0; y < Self->Bitmap->Height; y++) {
            i = 0;
            for (x=0; x < (Self->Bitmap->Width<<2); x+=4) {
               row[i++] = data[x+0];  /* Blue */
               row[i++] = data[x+1];  /* Green */
               row[i++] = data[x+2];  /* Red */
            }
            png_write_row(write_ptr, row_pointers);
            data += Self->Bitmap->LineWidth;
         }
      }
   }
   else if (Self->Bitmap->BytesPerPixel IS 2) {
      if (Self->Flags & PCF_ALPHA) {
         UBYTE row[Self->Bitmap->Width * 4];
         UWORD *data;
         UBYTE *mask;
         WORD maskx;

         row_pointers = row;
         data = (UWORD *)Self->Bitmap->Data;
         mask = Self->Mask->Data;
         for (y=0; y < Self->Bitmap->Height; y++) {
            i = 0;
            maskx = 0;
            for (x=0; x < Self->Bitmap->Width; x++) {
               row[i++] = UnpackBlue(Self->Bitmap, data[x]);
               row[i++] = UnpackGreen(Self->Bitmap, data[x]);
               row[i++] = UnpackRed(Self->Bitmap, data[x]);
               row[i++] = mask[maskx++];
            }
            png_write_row(write_ptr, row_pointers);
            data = (UWORD *)(((UBYTE *)data) + Self->Bitmap->LineWidth);
            mask += Self->Mask->LineWidth;
         }
      }
      else {
         UBYTE row[Self->Bitmap->Width * 3];
         UWORD *data;

         row_pointers = row;
         data = (UWORD *)Self->Bitmap->Data;
         for (y=0; y < Self->Bitmap->Height; y++) {
            i = 0;
            for (x=0; x < Self->Bitmap->Width; x++) {
               row[i++] = UnpackBlue(Self->Bitmap, data[x]);
               row[i++] = UnpackGreen(Self->Bitmap, data[x]);
               row[i++] = UnpackRed(Self->Bitmap, data[x]);
            }
            png_write_row(write_ptr, row_pointers);
            data = (UWORD *)(((UBYTE *)data) + Self->Bitmap->LineWidth);
         }
      }
   }

   png_write_end(write_ptr, NULL);

   error = ERR_Okay;

exit:
   png_destroy_write_struct(&write_ptr, &info_ptr);

   if ((Args) AND (Args->DestinationID)) {
      if (file) ReleaseObject(file);
   }
   else {
      if (file) Action(AC_Free, file, NULL);
   }

   StepBack();
   if (error) return(ObjectError(Self, ERH_SaveImage, error));
   else return(ERR_Okay);
}

/*****************************************************************************
** PNG: SaveToObject
*/

static ERROR PNG_SaveToObject(struct Picture *Self, struct acSaveToObject *Args)
{
   /*** Pass the action call directly on to SaveImage ***/
   return(ActionTags(AC_SaveImage, Self, Args->DestinationID, NULL));
}

/****************************************************************************/

static void read_row_callback(png_structp read_ptr, png_uint_32 row, int pass)
{

}

static void write_row_callback(png_structp write_ptr, png_uint_32 row, int pass)
{

}

/*****************************************************************************
** Read functions
*/

void png_read_data(png_structp png, png_bytep data, png_size_t length)
{
   struct acRead read;

   read.Buffer = data;
   read.Length = length;
   if ((Action(AC_Read, png->io_ptr, &read) != ERR_Okay) OR (read.Result != length)) {
      png_error(png, "File read error");
   }
}

void png_set_read_fn(png_structp png_ptr, png_voidp io_ptr, png_rw_ptr read_data_fn)
{
   png_ptr->io_ptr = io_ptr;
   png_ptr->read_data_fn = png_read_data;
   png_ptr->output_flush_fn = NULL;
}

/*****************************************************************************
** Write functions.
*/

void png_write_data(png_structp png, png_bytep data, png_size_t length)
{
   struct acWrite write;

   write.Buffer = data;
   write.Length = length;
   if ((Action(AC_Write, png->io_ptr, &write) != ERR_Okay) OR (write.Result != length)) {
      png_error(png, "File write error");
   }
}

void png_flush(png_structp png_ptr)
{

}

void png_set_write_fn(png_structp png_ptr, png_voidp io_ptr, png_rw_ptr write_data_fn, png_flush_ptr output_flush_fn)
{
   png_ptr->io_ptr = io_ptr;
   png_ptr->write_data_fn = png_write_data;
   png_ptr->output_flush_fn = NULL;
}

/*****************************************************************************
** Memory allocation routines.
*/

APTR png_create_struct(int type)
{
   LONG size;
   APTR struct_ptr;

   if (type IS PNG_STRUCT_INFO) size = sizeof(png_info);
   else if (type IS PNG_STRUCT_PNG) size = sizeof(png_struct);
   else return (NULL);

   if (AllocMemory(size, MEM_DATA, &struct_ptr, NULL) IS ERR_Okay) {
      return(struct_ptr);
   }
   else return(NULL);
}

void png_destroy_struct(APTR struct_ptr)
{
   if (struct_ptr != NULL) FreeMemory(struct_ptr);
}

APTR png_malloc(png_structp png_ptr, png_uint_32 size)
{
   APTR ret;

   if (!png_ptr) return(NULL);

   if (AllocMemory(size, MEM_DATA, &ret, NULL) IS ERR_Okay) {
      return(ret);
   }
   else {
      png_error(png_ptr, "Out of Memory");
      return(NULL);
   }
}

void png_free(png_structp png_ptr, APTR ptr)
{
   if ((!png_ptr) OR (!ptr)) return;
   FreeMemory(ptr);
}

APTR png_memcpy_check (png_structp png_ptr, APTR s1, APTR s2, png_uint_32 length)
{
   LONG size;

   size = length;
   if (size != length) png_error(png_ptr,"Overflow in png_memcpy_check.");
   return(png_memcpy (s1, s2, size));
}

APTR png_memset_check(png_structp png_ptr, APTR s1, int value, png_uint_32 length)
{
   png_size_t size;

   size = length;
   if (size != length) png_error(png_ptr,"Overflow in png_memset_check.");
   return(png_memset (s1, value, size));
}

/*****************************************************************************
** PNG Error Handling Functions
*/

void png_error(png_structp png_ptr, png_const_charp message)
{
   DPrintF("!PNG:","%s: %s", message);
   glError = TRUE;
}

void png_warning(png_structp png_ptr, png_const_charp message)
{
   DPrintF("@PNG:","Warning: %s", message);
   glError = FALSE;
}

/* These utilities are used internally to build an error message that relates
** to the current chunk.  The chunk name comes from png_ptr->chunk_name,
** this is used to prefix the message.  The message is limited in length
** to 63 bytes, the name characters are output as hex digits wrapped in []
** if the character is invalid.
*/

#define isnonalpha(c) ((c) < 41 || (c) > 122 || ((c) > 90 && (c) < 97))

static PNG_CONST char png_digit[16] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };

static void png_format_buffer(png_structp png_ptr, png_charp buffer, png_const_charp message)
{
   int iout = 0, iin = 0;

   while (iin < 4) {
      int c = png_ptr->chunk_name[iin++];
      if (isnonalpha(c)) {
         buffer[iout++] = '[';
         buffer[iout++] = png_digit[(c & 0xf0) >> 4];
         buffer[iout++] = png_digit[c & 0xf];
         buffer[iout++] = ']';
      } else {
         buffer[iout++] = c;
      }
   }

   if (message == NULL)
      buffer[iout] = 0;
   else {
      buffer[iout++] = ':';
      buffer[iout++] = ' ';
      png_memcpy(buffer+iout, message, 64);
      buffer[iout+63] = 0;
   }
}

void png_chunk_error(png_structp png_ptr, png_const_charp message)
{
   char msg[16+64];
   png_format_buffer(png_ptr, msg, message);
   png_error(png_ptr, msg);
}

void png_chunk_warning(png_structp png_ptr, png_const_charp message)
{
   char msg[16+64];
   png_format_buffer(png_ptr, msg, message);
   png_warning(png_ptr, msg);
}
